/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package edu.isi.wings.catalog.component.classes;

import edu.isi.kcap.ontapi.KBTriple;
import edu.isi.wings.workflow.template.classes.Role;
import edu.isi.wings.workflow.template.classes.sets.Binding;
import edu.isi.wings.workflow.template.classes.variables.*;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Class used during workflow generation to pass component information to and
 * from the component catalog to the workflow planner
 *
 * <ul>
 * <li>component : Component Variable
 * <li>roleMap : Component Input Role <=> Template Variable
 * <li>requirements : Also known as DOD's (Data Object Descriptions). These are
 * basically a list of KBTriple Objects
 * </ul>
 */
public class ComponentPacket implements Serializable {

  private static final long serialVersionUID = 1L;

  private ComponentVariable component;
  private LinkedHashMap<Role, Variable> roleMap;
  private ArrayList<KBTriple> requirements;

  // Keep a reverse mapping
  private LinkedHashMap<Variable, Role> variableMap;
  // Reasoner explanations (provided by component catalog)
  private HashSet<String> explanations;
  // If the reasoner marked this ComponentDetails packet as invalid
  public boolean isInvalid;
  // If the reasoner marked this ComponentDetails packet as a No-operation (i.e. skip processing)
  public boolean isNoOperation;
  // Mapping of compatible output roles to input roles to pass through
  private HashMap<String, String> noOperationIOPassthrough;

  private ArrayList<String> inputRoles;

  public ComponentPacket(
    ComponentVariable component,
    Map<Role, Variable> roleMap,
    ArrayList<KBTriple> requirements
  ) {
    this.component = component;
    this.roleMap = new LinkedHashMap<Role, Variable>(roleMap);
    this.variableMap = createReverseMap(roleMap);
    this.requirements = requirements;
    this.explanations = new HashSet<String>();
    this.inputRoles = new ArrayList<String>();
    this.noOperationIOPassthrough = new HashMap<String, String>();
    this.isInvalid = false;
  }

  public LinkedHashMap<Variable, Role> createReverseMap(
    Map<Role, Variable> map
  ) {
    LinkedHashMap<Variable, Role> rmap = new LinkedHashMap<Variable, Role>();
    for (Role cp : map.keySet()) {
      rmap.put(map.get(cp), cp);
    }
    return rmap;
  }

  public LinkedHashMap<Role, Variable> createMapFromReverseMap(
    Map<Variable, Role> rmap
  ) {
    LinkedHashMap<Role, Variable> map = new LinkedHashMap<Role, Variable>();
    for (Variable cp : rmap.keySet()) {
      map.put(rmap.get(cp), cp);
    }
    return map;
  }

  /**
   * Getter for property 'component'.
   *
   * @return Value for property 'component'.
   */
  public ComponentVariable getComponent() {
    return component;
  }

  /**
   * Setter for property 'component'.
   *
   * @param component
   *            Value to set for property 'component'.
   */
  public void setComponent(ComponentVariable component) {
    this.component = component;
  }

  /**
   * Getter for property 'roleMap'.
   *
   * @return Value for property 'roleMap'.
   */
  public LinkedHashMap<Role, Variable> getRoleMap() {
    return roleMap;
  }

  /**
   * Setter for property 'roleMap'.
   *
   * @param roleMap
   *            Value to set for property 'roleMap'.
   */
  public void setRoleMap(LinkedHashMap<Role, Variable> inputMaps) {
    this.roleMap = inputMaps;
    this.variableMap = createReverseMap(inputMaps);
  }

  /**
   * Getter for property 'requirements'.
   *
   * @return Value for property 'requirements'.
   */
  public ArrayList<KBTriple> getRequirements() {
    return requirements;
  }

  /**
   * Setter for property 'requirements'.
   *
   * @param requirements
   *            Value to set for property 'requirements'.
   */
  public void setRequirements(ArrayList<KBTriple> requirements) {
    this.requirements = requirements;
  }

  /**
   * Get Variable <=> Input Role Mappings
   *
   * @return A LinkedHashMap of Variable <=> Role objects
   */
  public LinkedHashMap<Variable, Role> getVariableMap() {
    return variableMap;
  }

  /**
   * Get Input Role ID <=> Variable Mappings
   *
   * @return Mapping Role IDs to Variable
   */
  public LinkedHashMap<String, Variable> getStringRoleMaps() {
    LinkedHashMap<String, Variable> map = new LinkedHashMap<String, Variable>();
    for (Role cp : roleMap.keySet()) {
      map.put(cp.getRoleId(), roleMap.get(cp));
    }
    return map;
  }

  /**
   * Get VarID <=> Input Role Mappings
   *
   * @return Mapping Variable IDs to Roles
   */
  public LinkedHashMap<String, Role> getStringVariableMap() {
    LinkedHashMap<String, Role> map = new LinkedHashMap<String, Role>();
    for (Variable var : variableMap.keySet()) {
      map.put(var.getID(), variableMap.get(var));
    }
    return map;
  }

  /**
   * Add Explanations
   *
   * @param explanations
   *            The reasoning explaining the contents of this CMR (usually
   *            returned from the Catalog)
   */
  public void addExplanations(HashSet<String> explanations) {
    this.explanations.addAll(explanations);
  }

  public void addExplanations(String explanation) {
    this.explanations.add(explanation);
  }

  /**
   * Set Roles to be inputs
   */
  public void setInputRoles(ArrayList<String> roleid) {
    this.inputRoles = roleid;
  }

  public void addInputRole(String roleid) {
    this.inputRoles.add(roleid);
  }

  /**
   * Retrieve roles as input or output
   */
  public boolean isInputRole(String roleid) {
    return this.inputRoles.contains(roleid);
  }

  /**
   * Get Explanations
   *
   * @return The reasoning explaining the contents of this CMR (usually
   *         returned from the Catalog)
   */
  public HashSet<String> getExplanations() {
    return this.explanations;
  }

  /**
   * Set Invalid Flag
   *
   * @param isInvalid
   *            Mark this CMR as invalid (i.e. not to be used)
   */
  public void setInvalidFlag(boolean isInvalid) {
    this.isInvalid = isInvalid;
  }

  /**
   * Get Invalid Flag
   *
   * @return The reasoning explaining the contents of this CMR (usually
   *         returned from the Catalog)
   */
  public boolean getInvalidFlag() {
    return this.isInvalid;
  }

  public boolean getNoOperationFlag() {
    return isNoOperation;
  }

  public void setNoOperationFlag(boolean isNoOperation) {
    this.isNoOperation = isNoOperation;
  }

  public HashMap<String, String> getNoOperationIOPassthrough() {
    return noOperationIOPassthrough;
  }

  public void addNoOperationIOPassthrough(String invar, String outvar) {
    this.noOperationIOPassthrough.put(invar, outvar);
  }

  @Override
  public String toString() {
    return (
      "ComponentDetails{invalid=" +
      this.isInvalid +
      ",component=" +
      component +
      ", roleMap=" +
      roleMap +
      ", requirements=" +
      requirements +
      '}'
    );
  }

  public ComponentPacket clone() {
    // Clone the packet
    // We just need to have separate variable bindings
    HashMap<Role, Variable> roleMap = this.getRoleMap();
    HashMap<Role, Variable> sendMap = new HashMap<Role, Variable>();

    for (Role r : roleMap.keySet()) {
      Variable var = roleMap.get(r);
      Variable sendVar = new Variable(var.getID(), var.getVariableType());
      sendVar.setBreakpoint(var.isBreakpoint());
      if (var.getBinding() != null) sendVar.setBinding(
        (Binding) var.getBinding().clone()
      );
      sendMap.put(r, sendVar);
    }
    ComponentPacket pthis = new ComponentPacket(
      this.getComponent(),
      sendMap,
      this.getRequirements()
    );
    pthis.addExplanations(this.getExplanations());
    pthis.setInvalidFlag(this.getInvalidFlag());
    return pthis;
  }
}
